@@ -1,39 +1,94 @@
 //! @file a68g.c
 //! @author J. Marcel van der Veer
-//
+
 //! @section Copyright
-//
-// This file is part of Algol68G - an Algol 68 compiler-interpreter.
-// Copyright 2001-2023 J. Marcel van der Veer <algol68g@xs4all.nl>.
-//
+//!
+//! This file is part of Algol68G - an Algol 68 compiler-interpreter.
+//! Copyright 2001-2024 J. Marcel van der Veer [algol68g@xs4all.nl].
+
 //! @section License
-//
-// This program is free software; you can redistribute it and/or modify it 
-// under the terms of the GNU General Public License as published by the 
-// Free Software Foundation; either version 3 of the License, or 
-// (at your option) any later version.
-//
-// This program is distributed in the hope that it will be useful, but 
-// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
-// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for 
-// more details. You should have received a copy of the GNU General Public 
-// License along with this program. If not, see <http://www.gnu.org/licenses/>.
+//!
+//! This program is free software; you can redistribute it and/or modify it 
+//! under the terms of the GNU General Public License as published by the 
+//! Free Software Foundation; either version 3 of the License, or 
+//! (at your option) any later version.
+//!
+//! This program is distributed in the hope that it will be useful, but 
+//! WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
+//! or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for 
+//! more details. You should have received a copy of the GNU General Public 
+//! License along with this program. If not, see [http://www.gnu.org/licenses/].
+
+//! @section Synopsis
+//!
+//! Algol 68 Genie main driver.
+
+// --assertions, --noassertions, switch elaboration of assertions on or off.
+// --backtrace, --nobacktrace, switch stack backtracing in case of a runtime error.
+// --boldstropping, set stropping mode to bold stropping.
+// --brackets, consider [ .. ] and { .. } as equivalent to ( .. ).
+// --check, --norun, check syntax only, interpreter does not start.
+// --clock, report execution time excluding compilation time.
+// --compile, compile source file.
+// --debug, --monitor, start execution in the debugger and debug in case of runtime error.
+// --echo string, echo 'string' to standard output.
+// --execute unit, execute algol 68 unit 'unit'.
+// --exit, --, ignore next options.
+// --extensive, make extensive listing.
+// --file string, accept string as generic filename.
+// --frame 'number', set frame stack size to 'number'.
+// --handles 'number', set handle space size to 'number'.
+// --heap 'number', set heap size to 'number'.
+// --keep, --nokeep, switch object file deletion off or on.
+// --listing, make concise listing.
+// --moids, make overview of moids in listing file.
+// -O0, -O1, -O2, -O3, switch compilation on and pass option to back-end C compiler.
+// --optimise, --nooptimise, switch compilation on or off.
+// --pedantic, equivalent to --warnings --portcheck.
+// --portcheck, --noportcheck, switch portability warnings on or off.
+// --pragmats, --nopragmats, switch elaboration of pragmat items on or off.
+// --precision 'number', set precision for long long modes to 'number' significant digits.
+// --preludelisting, make a listing of preludes.
+// --pretty-print, pretty-print the source file.
+// --print unit, print value yielded by algol 68 unit 'unit'.
+// --quiet, suppresses all warning diagnostics.
+// --quotestropping, set stropping mode to quote stropping.
+// --reductions, print parser reductions.
+// --run, override --check/--norun options.
+// --rerun, run using already compiled code.
+// --script, set next option as source file name; pass further options to algol 68 program.
+// --source, --nosource, switch listing of source lines in listing file on or off.
+// --stack 'number', set expression stack size to 'number'.
+// --statistics, print statistics in listing file.
+// --strict, disable most extensions to Algol 68 syntax.
+// --timelimit 'number', interrupt the interpreter after 'number' seconds.
+// --trace, --notrace, switch tracing of a running program on or off.
+// --tree, --notree, switch syntax tree listing in listing file on or off.
+// --unused, make an overview of unused tags in the listing file.
+// --verbose, inform on program actions.
+// --version, state version of the running copy.
+// --warnings, --nowarnings, switch warning diagnostics on or off.
+// --xref, --noxref, switch cross reference in the listing file on or off.
 
 #include "a68g.h"
-#include "a68g-prelude.h"
-#include "a68g-mp.h"
-#include "a68g-genie.h"
-#include "a68g-prelude-mathlib.h"
-#include "a68g-postulates.h"
-#include "a68g-parser.h"
-#include "a68g-options.h"
-#include "a68g-optimiser.h"
 #include "a68g-listing.h"
+#include "a68g-mp.h"
+#include "a68g-optimiser.h"
+#include "a68g-options.h"
+#include "a68g-parser.h"
+#include "a68g-postulates.h"
+#include "a68g-genie.h"
+#include "a68g-prelude.h"
+#include "a68g-prelude-mathlib.h"
 
 #if defined (HAVE_MATHLIB)
 #include <Rmath.h>
 #endif
 
+#if defined (HAVE_CURL)
+#include <curl/curl.h>
+#endif
+
 GLOBALS_T common;
 
 #define EXTENSIONS 11
@@ -64,26 +119,25 @@ void verbosity (void)
 void state_license (FILE_T f)
 {
 #define PR(s)\
-  ASSERT (snprintf(A68 (output_line), SNPRINTF_SIZE, "%s\n", (s)) >= 0);\
+  ASSERT (a68_bufprt (A68 (output_line), SNPRINTF_SIZE, "%s\n", (s)) >= 0);\
   WRITE (f, A68 (output_line));
-//
-  if (f == STDOUT_FILENO) {
+  if (f == A68_STDOUT) {
     io_close_tty_line ();
   }
-  ASSERT (snprintf (A68 (output_line), SNPRINTF_SIZE, "Algol 68 Genie %s\n", PACKAGE_VERSION) >= 0);
+  ASSERT (a68_bufprt (A68 (output_line), SNPRINTF_SIZE, "Algol 68 Genie %s\n", PACKAGE_VERSION) >= 0);
   WRITE (f, A68 (output_line));
-  ASSERT (snprintf (A68 (output_line), SNPRINTF_SIZE, "Copyright 2001-2023 %s.\n", PACKAGE_BUGREPORT) >= 0);
+  ASSERT (a68_bufprt (A68 (output_line), SNPRINTF_SIZE, "Copyright 2001-2024 %s.\n", PACKAGE_BUGREPORT) >= 0);
   WRITE (f, A68 (output_line));
   PR ("");
-  ASSERT (snprintf (A68 (output_line), SNPRINTF_SIZE, "This is free software covered by the GNU General Public License.\n") >= 0);
+  ASSERT (a68_bufprt (A68 (output_line), SNPRINTF_SIZE, "This is free software covered by the GNU General Public License.\n") >= 0);
   WRITE (f, A68 (output_line));
-  ASSERT (snprintf (A68 (output_line), SNPRINTF_SIZE, "There is ABSOLUTELY NO WARRANTY for Algol 68 Genie;\n") >= 0);
+  ASSERT (a68_bufprt (A68 (output_line), SNPRINTF_SIZE, "There is ABSOLUTELY NO WARRANTY for Algol 68 Genie;\n") >= 0);
   WRITE (f, A68 (output_line));
-  ASSERT (snprintf (A68 (output_line), SNPRINTF_SIZE, "not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n") >= 0);
+  ASSERT (a68_bufprt (A68 (output_line), SNPRINTF_SIZE, "not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n") >= 0);
   WRITE (f, A68 (output_line));
   PR ("See the GNU General Public License for more details.");
   PR ("");
-  ASSERT (snprintf (A68 (output_line), SNPRINTF_SIZE, "Please report bugs to %s.\n", PACKAGE_BUGREPORT) >= 0);
+  ASSERT (a68_bufprt (A68 (output_line), SNPRINTF_SIZE, "Please report bugs to %s.\n", PACKAGE_BUGREPORT) >= 0);
   WRITE (f, A68 (output_line));
 #undef PR
 }
@@ -93,79 +147,85 @@ void state_license (FILE_T f)
 void state_version (FILE_T f)
 {
 #define PR(s)\
-  ASSERT (snprintf(A68 (output_line), SNPRINTF_SIZE, "%s\n", (s)) >= 0);\
+  ASSERT (a68_bufprt (A68 (output_line), SNPRINTF_SIZE, "%s\n", (s)) >= 0);\
   WRITE (f, A68 (output_line));
-//
-  if (f == STDOUT_FILENO) {
+  if (f == A68_STDOUT) {
     io_close_tty_line ();
   }
   state_license (f);
   PR ("");
 #if defined (BUILD_WIN32)
-  ASSERT (snprintf (A68 (output_line), SNPRINTF_SIZE, "WIN32 executable\n") >= 0);
+  ASSERT (a68_bufprt (A68 (output_line), SNPRINTF_SIZE, "WIN32 executable\n") >= 0);
   WRITE (f, A68 (output_line));
   WRITELN (f, "");
 #endif
-//ASSERT (snprintf (A68 (output_line), SNPRINTF_SIZE, "Algol 68 Genie version %d\n", A68_LEVEL) >= 0);
-//WRITE (f, A68 (output_line));
 #if (A68_LEVEL >= 3)
-  ASSERT (snprintf (A68 (output_line), SNPRINTF_SIZE, "With hardware support for long modes\n") >= 0);
+  ASSERT (a68_bufprt (A68 (output_line), SNPRINTF_SIZE, "With hardware support for long modes\n") >= 0);
   WRITE (f, A68 (output_line));
 #endif
 #if defined (BUILD_A68_COMPILER)
-  ASSERT (snprintf (A68 (output_line), SNPRINTF_SIZE, "With compilation support\n") >= 0);
+  ASSERT (a68_bufprt (A68 (output_line), SNPRINTF_SIZE, "With plugin-compilation support\n") >= 0);
   WRITE (f, A68 (output_line));
-#if defined (C_COMPILER)
-  ASSERT (snprintf (A68 (output_line), SNPRINTF_SIZE, "     C compiler is %s\n", C_COMPILER) >= 0);
-  WRITE (f, A68 (output_line));
-#endif
 #endif
 #if defined (BUILD_PARALLEL_CLAUSE)
-  ASSERT (snprintf (A68 (output_line), SNPRINTF_SIZE, "With parallel-clause support\n") >= 0);
-  WRITE (f, A68 (output_line));
-#endif
-#if defined (HAVE_GNU_MPFR)
-  ASSERT (snprintf (A68 (output_line), SNPRINTF_SIZE, "With GNU MP %s\n", gmp_version) >= 0);
-  WRITE (f, A68 (output_line));
-  ASSERT (snprintf (A68 (output_line), SNPRINTF_SIZE, "With GNU MPFR %s\n", mpfr_get_version ()) >= 0);
-  WRITE (f, A68 (output_line));
-#endif
-#if defined (HAVE_MATHLIB)
-  ASSERT (snprintf (A68 (output_line), SNPRINTF_SIZE, "With mathlib from R %s\n", R_VERSION_STRING) >= 0);
-  WRITE (f, A68 (output_line));
-#endif
-#if defined (HAVE_GSL)
-  ASSERT (snprintf (A68 (output_line), SNPRINTF_SIZE, "With GNU Scientific Library %s\n", GSL_VERSION) >= 0);
-  WRITE (f, A68 (output_line));
-#endif
-#if defined (HAVE_GNU_PLOTUTILS)
-  ASSERT (snprintf (A68 (output_line), SNPRINTF_SIZE, "With GNU plotutils %s\n", PL_LIBPLOT_VER_STRING) >= 0);
-  WRITE (f, A68 (output_line));
-#endif
-#if defined (HAVE_CURSES)
-  ASSERT (snprintf (A68 (output_line), SNPRINTF_SIZE, "With curses %s\n", NCURSES_VERSION) >= 0);
-  WRITE (f, A68 (output_line));
-#endif
-#if defined (BUILD_HTTP)
-  ASSERT (snprintf (A68 (output_line), SNPRINTF_SIZE, "With TCP/IP support\n") >= 0);
+  ASSERT (a68_bufprt (A68 (output_line), SNPRINTF_SIZE, "With parallel-clause support\n") >= 0);
   WRITE (f, A68 (output_line));
 #endif
 #if defined (HAVE_POSTGRESQL)
-  ASSERT (snprintf (A68 (output_line), SNPRINTF_SIZE, "With PostgreSQL support\n") >= 0);
+  ASSERT (a68_bufprt (A68 (output_line), SNPRINTF_SIZE, "With PostgreSQL support\n") >= 0);
+  WRITE (f, A68 (output_line));
+#endif
+#if defined (HAVE_CURL)
+  curl_version_info_data *data = curl_version_info(CURLVERSION_NOW);
+  ASSERT (a68_bufprt (A68 (output_line), SNPRINTF_SIZE, "With curl %s\n", data->version) >= 0);
+  WRITE (f, A68 (output_line));
+#endif
+#if defined (HAVE_GNU_MPFR)
+  ASSERT (a68_bufprt (A68 (output_line), SNPRINTF_SIZE, "With GNU MP %s\n", gmp_version) >= 0);
+  WRITE (f, A68 (output_line));
+  ASSERT (a68_bufprt (A68 (output_line), SNPRINTF_SIZE, "With GNU MPFR %s\n", mpfr_get_version ()) >= 0);
+  WRITE (f, A68 (output_line));
+#endif
+#if defined (HAVE_MATHLIB)
+  ASSERT (a68_bufprt (A68 (output_line), SNPRINTF_SIZE, "With mathlib from R %s\n", R_VERSION_STRING) >= 0);
+  WRITE (f, A68 (output_line));
+#endif
+#if defined (HAVE_GSL)
+  ASSERT (a68_bufprt (A68 (output_line), SNPRINTF_SIZE, "With GNU Scientific Library %s\n", GSL_VERSION) >= 0);
+  WRITE (f, A68 (output_line));
+#endif
+#if defined (HAVE_GNU_PLOTUTILS)
+  ASSERT (a68_bufprt (A68 (output_line), SNPRINTF_SIZE, "With GNU plotutils %s\n", PL_LIBPLOT_VER_STRING) >= 0);
+  WRITE (f, A68 (output_line));
+#endif
+#if defined (HAVE_CURSES)
+  #if defined (NCURSES_VERSION)
+    ASSERT (a68_bufprt (A68 (output_line), SNPRINTF_SIZE, "With ncurses %s\n", NCURSES_VERSION) >= 0);
+  #else
+    ASSERT (a68_bufprt (A68 (output_line), SNPRINTF_SIZE, "With curses support\n") >= 0);
+  #endif
   WRITE (f, A68 (output_line));
 #endif
 #if defined (_CS_GNU_LIBC_VERSION) && defined (BUILD_UNIX)
   if (confstr (_CS_GNU_LIBC_VERSION, A68 (input_line), BUFFER_SIZE) > (size_t) 0) {
-    ASSERT (snprintf (A68 (output_line), SNPRINTF_SIZE, "GNU libc version %s\n", A68 (input_line)) >= 0);
+    ASSERT (a68_bufprt (A68 (output_line), SNPRINTF_SIZE, "GNU libc version %s\n", A68 (input_line)) >= 0);
     WRITE (f, A68 (output_line));
   }
 #if (defined (BUILD_PARALLEL_CLAUSE) && defined (_CS_GNU_LIBPTHREAD_VERSION))
   if (confstr (_CS_GNU_LIBPTHREAD_VERSION, A68 (input_line), BUFFER_SIZE) > (size_t) 0) {
-    ASSERT (snprintf (A68 (output_line), SNPRINTF_SIZE, "GNU libpthread version %s\n", A68 (input_line)) >= 0);
+    ASSERT (a68_bufprt (A68 (output_line), SNPRINTF_SIZE, "GNU libpthread version %s\n", A68 (input_line)) >= 0);
     WRITE (f, A68 (output_line));
   }
 #endif
 #endif
+#define RSIZE(n) (unt) (sizeof (n) / sizeof (int))
+#if defined (BUILD_A68_COMPILER) && defined (C_COMPILER)
+  ASSERT (a68_bufprt (A68 (output_line), SNPRINTF_SIZE, "Build level %d.%x%x%x%x %s %s\n", A68_LEVEL, RSIZE (INT_T), RSIZE (REAL_T), RSIZE (MP_INT_T), RSIZE (MP_REAL_T), C_COMPILER, __DATE__) >= 0);
+#else
+  ASSERT (a68_bufprt (A68 (output_line), SNPRINTF_SIZE, "Build level %d.%x%x%x%x %s\n", A68_LEVEL, RSIZE (INT_T), RSIZE (REAL_T), RSIZE (MP_INT_T), RSIZE (MP_REAL_T), __DATE__) >= 0);
+#endif
+#undef RSIZE
+WRITE (f, A68 (output_line));
 #undef PR
 }
 
@@ -173,13 +233,13 @@ void state_version (FILE_T f)
 
 void online_help (FILE_T f)
 {
-  if (f == STDOUT_FILENO) {
+  if (f == A68_STDOUT) {
     io_close_tty_line ();
   }
   state_license (f);
-  ASSERT (snprintf (A68 (output_line), SNPRINTF_SIZE, "Usage: %s [options | filename]", A68 (a68_cmd_name)) >= 0);
+  ASSERT (a68_bufprt (A68 (output_line), SNPRINTF_SIZE, "Usage: %s [options | filename]", A68 (a68_cmd_name)) >= 0);
   WRITELN (f, A68 (output_line));
-  ASSERT (snprintf (A68 (output_line), SNPRINTF_SIZE, "For help: %s --apropos [keyword]", A68 (a68_cmd_name)) >= 0);
+  ASSERT (a68_bufprt (A68 (output_line), SNPRINTF_SIZE, "For help: %s --apropos [keyword]", A68 (a68_cmd_name)) >= 0);
   WRITELN (f, A68 (output_line));
 }
 
@@ -188,9 +248,9 @@ void online_help (FILE_T f)
 void announce_phase (char *t)
 {
   if (OPTION_VERBOSE (&A68_JOB)) {
-    ASSERT (snprintf (A68 (output_line), SNPRINTF_SIZE, "%s: %s", A68 (a68_cmd_name), t) >= 0);
+    ASSERT (a68_bufprt (A68 (output_line), SNPRINTF_SIZE, "%s: %s", A68 (a68_cmd_name), t) >= 0);
     io_close_tty_line ();
-    WRITE (STDOUT_FILENO, A68 (output_line));
+    WRITE (A68_STDOUT, A68 (output_line));
   }
 }
 
@@ -204,7 +264,7 @@ BOOL_T strip_extension (char *ext)
   int nlen = (int) strlen (FILE_SOURCE_NAME (&A68_JOB)), xlen = (int) strlen (ext);
   if (nlen > xlen && strcmp (&(FILE_SOURCE_NAME (&A68_JOB)[nlen - xlen]), ext) == 0) {
     char *fn = (char *) get_heap_space ((size_t) (nlen + 1));
-    bufcpy (fn, FILE_SOURCE_NAME (&A68_JOB), nlen);
+    a68_bufcpy (fn, FILE_SOURCE_NAME (&A68_JOB), nlen);
     fn[nlen - xlen] = NULL_CHAR;
     a68_free (FILE_GENERIC_NAME (&A68_JOB));
     FILE_GENERIC_NAME (&A68_JOB) = new_string (fn, NO_TEXT);
@@ -219,31 +279,29 @@ BOOL_T strip_extension (char *ext)
 
 void open_with_extensions (void)
 {
-  int k;
   FILE_SOURCE_FD (&A68_JOB) = -1;
-  for (k = 0; k < EXTENSIONS && FILE_SOURCE_FD (&A68_JOB) == -1; k++) {
+  for (int k = 0; k < EXTENSIONS && FILE_SOURCE_FD (&A68_JOB) == -1; k++) {
     int len;
     char *fn = NULL;
     if (extensions[k] == NO_TEXT) {
       len = (int) strlen (FILE_INITIAL_NAME (&A68_JOB)) + 1;
       fn = (char *) get_heap_space ((size_t) len);
-      bufcpy (fn, FILE_INITIAL_NAME (&A68_JOB), len);
+      a68_bufcpy (fn, FILE_INITIAL_NAME (&A68_JOB), len);
     } else {
       len = (int) strlen (FILE_INITIAL_NAME (&A68_JOB)) + (int) strlen (extensions[k]) + 1;
       fn = (char *) get_heap_space ((size_t) len);
-      bufcpy (fn, FILE_INITIAL_NAME (&A68_JOB), len);
-      bufcat (fn, extensions[k], len);
+      a68_bufcpy (fn, FILE_INITIAL_NAME (&A68_JOB), len);
+      a68_bufcat (fn, extensions[k], len);
     }
     FILE_SOURCE_FD (&A68_JOB) = open (fn, O_RDONLY | O_BINARY);
     if (FILE_SOURCE_FD (&A68_JOB) != -1) {
-      int l;
       BOOL_T cont = A68_TRUE;
       a68_free (FILE_SOURCE_NAME (&A68_JOB));
       a68_free (FILE_GENERIC_NAME (&A68_JOB));
       FILE_SOURCE_NAME (&A68_JOB) = new_string (fn, NO_TEXT);
       FILE_GENERIC_NAME (&A68_JOB) = new_string (a68_basename (fn), NO_TEXT);
       FILE_PATH (&A68_JOB) = new_string (a68_dirname (fn), NO_TEXT);
-      for (l = 0; l < EXTENSIONS && cont; l++) {
+      for (int l = 0; l < EXTENSIONS && cont; l++) {
         if (strip_extension (extensions[l])) {
           cont = A68_FALSE;
         }
@@ -269,10 +327,7 @@ void a68_rm (char *fn)
 
 void compiler_interpreter (void)
 {
-  int len, num;
-#if defined (BUILD_A68_COMPILER)
   BOOL_T emitted = A68_FALSE;
-#endif
   TREE_LISTING_SAFE (&A68_JOB) = A68_FALSE;
   CROSS_REFERENCE_SAFE (&A68_JOB) = A68_FALSE;
   A68 (in_execution) = A68_FALSE;
@@ -307,8 +362,8 @@ void compiler_interpreter (void)
   SCAN_ERROR (FILE_INITIAL_NAME (&A68_JOB) == NO_TEXT, NO_LINE, NO_TEXT, ERROR_NO_SOURCE_FILE);
   FILE_BINARY_OPENED (&A68_JOB) = A68_FALSE;
   FILE_BINARY_WRITEMOOD (&A68_JOB) = A68_TRUE;
-  FILE_LIBRARY_OPENED (&A68_JOB) = A68_FALSE;
-  FILE_LIBRARY_WRITEMOOD (&A68_JOB) = A68_TRUE;
+  FILE_PLUGIN_OPENED (&A68_JOB) = A68_FALSE;
+  FILE_PLUGIN_WRITEMOOD (&A68_JOB) = A68_TRUE;
   FILE_LISTING_OPENED (&A68_JOB) = A68_FALSE;
   FILE_LISTING_WRITEMOOD (&A68_JOB) = A68_TRUE;
   FILE_OBJECT_OPENED (&A68_JOB) = A68_FALSE;
@@ -343,35 +398,35 @@ void compiler_interpreter (void)
   ABEND (FILE_SOURCE_NAME (&A68_JOB) == NO_TEXT, ERROR_INTERNAL_CONSISTENCY, __func__);
   ABEND (FILE_GENERIC_NAME (&A68_JOB) == NO_TEXT, ERROR_INTERNAL_CONSISTENCY, __func__);
 // Object file.
-  len = 1 + (int) strlen (FILE_GENERIC_NAME (&A68_JOB)) + (int) strlen (OBJECT_EXTENSION);
+  int len = 1 + (int) strlen (FILE_GENERIC_NAME (&A68_JOB)) + (int) strlen (OBJECT_EXTENSION);
   FILE_OBJECT_NAME (&A68_JOB) = (char *) get_heap_space ((size_t) len);
-  bufcpy (FILE_OBJECT_NAME (&A68_JOB), FILE_GENERIC_NAME (&A68_JOB), len);
-  bufcat (FILE_OBJECT_NAME (&A68_JOB), OBJECT_EXTENSION, len);
+  a68_bufcpy (FILE_OBJECT_NAME (&A68_JOB), FILE_GENERIC_NAME (&A68_JOB), len);
+  a68_bufcat (FILE_OBJECT_NAME (&A68_JOB), OBJECT_EXTENSION, len);
 // Binary.
-  len = 1 + (int) strlen (FILE_GENERIC_NAME (&A68_JOB)) + (int) strlen (LIBRARY_EXTENSION);
+  len = 1 + (int) strlen (FILE_GENERIC_NAME (&A68_JOB)) + (int) strlen (PLUGIN_EXTENSION);
   FILE_BINARY_NAME (&A68_JOB) = (char *) get_heap_space ((size_t) len);
-  bufcpy (FILE_BINARY_NAME (&A68_JOB), FILE_GENERIC_NAME (&A68_JOB), len);
-  bufcat (FILE_BINARY_NAME (&A68_JOB), BINARY_EXTENSION, len);
+  a68_bufcpy (FILE_BINARY_NAME (&A68_JOB), FILE_GENERIC_NAME (&A68_JOB), len);
+  a68_bufcat (FILE_BINARY_NAME (&A68_JOB), BINARY_EXTENSION, len);
 // Library file.
-  len = 1 + (int) strlen (FILE_GENERIC_NAME (&A68_JOB)) + (int) strlen (LIBRARY_EXTENSION);
-  FILE_LIBRARY_NAME (&A68_JOB) = (char *) get_heap_space ((size_t) len);
-  bufcpy (FILE_LIBRARY_NAME (&A68_JOB), FILE_GENERIC_NAME (&A68_JOB), len);
-  bufcat (FILE_LIBRARY_NAME (&A68_JOB), LIBRARY_EXTENSION, len);
+  len = 1 + (int) strlen (FILE_GENERIC_NAME (&A68_JOB)) + (int) strlen (PLUGIN_EXTENSION);
+  FILE_PLUGIN_NAME (&A68_JOB) = (char *) get_heap_space ((size_t) len);
+  a68_bufcpy (FILE_PLUGIN_NAME (&A68_JOB), FILE_GENERIC_NAME (&A68_JOB), len);
+  a68_bufcat (FILE_PLUGIN_NAME (&A68_JOB), PLUGIN_EXTENSION, len);
 // Listing file.
   len = 1 + (int) strlen (FILE_GENERIC_NAME (&A68_JOB)) + (int) strlen (LISTING_EXTENSION);
   FILE_LISTING_NAME (&A68_JOB) = (char *) get_heap_space ((size_t) len);
-  bufcpy (FILE_LISTING_NAME (&A68_JOB), FILE_GENERIC_NAME (&A68_JOB), len);
-  bufcat (FILE_LISTING_NAME (&A68_JOB), LISTING_EXTENSION, len);
+  a68_bufcpy (FILE_LISTING_NAME (&A68_JOB), FILE_GENERIC_NAME (&A68_JOB), len);
+  a68_bufcat (FILE_LISTING_NAME (&A68_JOB), LISTING_EXTENSION, len);
 // Pretty file.
   len = 1 + (int) strlen (FILE_GENERIC_NAME (&A68_JOB)) + (int) strlen (PRETTY_EXTENSION);
   FILE_PRETTY_NAME (&A68_JOB) = (char *) get_heap_space ((size_t) len);
-  bufcpy (FILE_PRETTY_NAME (&A68_JOB), FILE_GENERIC_NAME (&A68_JOB), len);
-  bufcat (FILE_PRETTY_NAME (&A68_JOB), PRETTY_EXTENSION, len);
+  a68_bufcpy (FILE_PRETTY_NAME (&A68_JOB), FILE_GENERIC_NAME (&A68_JOB), len);
+  a68_bufcat (FILE_PRETTY_NAME (&A68_JOB), PRETTY_EXTENSION, len);
 // Script file.
   len = 1 + (int) strlen (FILE_GENERIC_NAME (&A68_JOB)) + (int) strlen (SCRIPT_EXTENSION);
   FILE_SCRIPT_NAME (&A68_JOB) = (char *) get_heap_space ((size_t) len);
-  bufcpy (FILE_SCRIPT_NAME (&A68_JOB), FILE_GENERIC_NAME (&A68_JOB), len);
-  bufcat (FILE_SCRIPT_NAME (&A68_JOB), SCRIPT_EXTENSION, len);
+  a68_bufcpy (FILE_SCRIPT_NAME (&A68_JOB), FILE_GENERIC_NAME (&A68_JOB), len);
+  a68_bufcat (FILE_SCRIPT_NAME (&A68_JOB), SCRIPT_EXTENSION, len);
 // Parser.
   a68_parser ();
   if (TOP_NODE (&A68_JOB) == NO_NODE) {
@@ -386,128 +441,51 @@ void compiler_interpreter (void)
   }
 // Finalise syntax tree.
   if (ERROR_COUNT (&A68_JOB) == 0) {
-    num = 0;
+    int num = 0;
     renumber_nodes (TOP_NODE (&A68_JOB), &num);
     NEST (TABLE (TOP_NODE (&A68_JOB))) = A68 (symbol_table_count) = 3;
     reset_symbol_table_nest_count (TOP_NODE (&A68_JOB));
     verbosity ();
   }
-//
   if (A68_MP (varying_mp_digits) > width_to_mp_digits (MP_MAX_DECIMALS)) {
     diagnostic (A68_WARNING, NO_NODE, WARNING_PRECISION, NO_LINE, 0, A68_MP (varying_mp_digits) * LOG_MP_RADIX);
   }
-// Compiler.
+// Plugin code generation and compilation.
   if (ERROR_COUNT (&A68_JOB) == 0 && OPTION_OPT_LEVEL (&A68_JOB) > NO_OPTIMISE) {
-    announce_phase ("optimiser (code generator)");
-    num = 0;
-    renumber_nodes (TOP_NODE (&A68_JOB), &num);
-    A68 (node_register) = (NODE_T **) get_heap_space ((size_t) num * sizeof (NODE_T));
-    ABEND (A68 (node_register) == NO_VAR, ERROR_ACTION, __func__);
-    register_nodes (TOP_NODE (&A68_JOB));
-    FILE_OBJECT_FD (&A68_JOB) = open (FILE_OBJECT_NAME (&A68_JOB), O_WRONLY | O_CREAT | O_TRUNC, A68_PROTECTION);
-    ABEND (FILE_OBJECT_FD (&A68_JOB) == -1, ERROR_ACTION, FILE_OBJECT_NAME (&A68_JOB));
-    FILE_OBJECT_OPENED (&A68_JOB) = A68_TRUE;
-    compiler (FILE_OBJECT_FD (&A68_JOB));
-    ASSERT (close (FILE_OBJECT_FD (&A68_JOB)) == 0);
-    FILE_OBJECT_OPENED (&A68_JOB) = A68_FALSE;
+    announce_phase ("plugin-compiler");
+    plugin_driver_code ();
 #if defined (BUILD_A68_COMPILER)
     emitted = A68_TRUE;
-#endif
-  }
-#if defined (BUILD_A68_COMPILER)
-// Only compile C if the A68 compiler found no errors (constant folder for instance).
-  if (ERROR_COUNT (&A68_JOB) == 0 && OPTION_OPT_LEVEL (&A68_JOB) > 0 && !OPTION_RUN_SCRIPT (&A68_JOB)) {
-    char cmd[BUFFER_SIZE], options[BUFFER_SIZE];
-    if (OPTION_RERUN (&A68_JOB) == A68_FALSE) {
-      announce_phase ("optimiser (code compiler)");
-      errno = 0;
-//
-// Compilation on Linux, BSD.
-// Build shared library using gcc or clang.
-// TODO: One day this should be all portable between platforms.
-//
-// -fno-stack-protector is needed for OS's that enforce -fstack-protector-strong which may give
-// undefined reference to `__stack_chk_fail_local'
-// by ld. Ubuntu is one such.
-//
-      ASSERT (snprintf (options, SNPRINTF_SIZE, "%s %s", optimisation_option (), A68_GCC_OPTIONS) >= 0);
-#if defined (HAVE_PIC)
-      bufcat (options, " ", BUFFER_SIZE);
-      bufcat (options, HAVE_PIC, BUFFER_SIZE);
-#endif
-      ASSERT (snprintf (cmd, SNPRINTF_SIZE, "%s -I%s %s -c -o \"%s\" \"%s\"", C_COMPILER, INCLUDEDIR, options, FILE_BINARY_NAME (&A68_JOB), FILE_OBJECT_NAME (&A68_JOB)) >= 0);
-      ABEND (system (cmd) != 0, ERROR_ACTION, cmd);
-      ASSERT (snprintf (cmd, SNPRINTF_SIZE, "ld -export-dynamic -shared -o \"%s\" \"%s\"", FILE_LIBRARY_NAME (&A68_JOB), FILE_BINARY_NAME (&A68_JOB)) >= 0);
-      ABEND (system (cmd) != 0, ERROR_ACTION, cmd);
-      a68_rm (FILE_BINARY_NAME (&A68_JOB));
+    if (ERROR_COUNT (&A68_JOB) == 0 && !OPTION_RUN_SCRIPT (&A68_JOB)) {
+      plugin_driver_compile ();
     }
     verbosity ();
-  }
 #else
-  if (OPTION_OPT_LEVEL (&A68_JOB) > 0) {
+    emitted = A68_FALSE;
     diagnostic (A68_WARNING | A68_FORCE_DIAGNOSTICS, TOP_NODE (&A68_JOB), WARNING_OPTIMISATION);
-  }
 #endif
+  }
 // Indenter.
   if (ERROR_COUNT (&A68_JOB) == 0 && OPTION_PRETTY (&A68_JOB)) {
     announce_phase ("indenter");
     indenter (&A68_JOB);
     verbosity ();
   }
-// Interpreter.
+// Interpreter initialisation.
   diagnostics_to_terminal (TOP_LINE (&A68_JOB), A68_ALL_DIAGNOSTICS);
-// Restore seed for rng.
-  GetRNGstate ();
-  A68 (f_entry) = TOP_NODE (&A68_JOB);
-//
-  if (ERROR_COUNT (&A68_JOB) == 0 && OPTION_COMPILE (&A68_JOB) == A68_FALSE && (OPTION_CHECK_ONLY (&A68_JOB) ? OPTION_RUN (&A68_JOB) : A68_TRUE)) {
-#if defined (BUILD_A68_COMPILER)
-    void *compile_lib;
-#endif
-    A68 (close_tty_on_exit) = A68_FALSE;        // Assuming no runtime errors a priori
-#if defined (BUILD_A68_COMPILER)
-    if (OPTION_RUN_SCRIPT (&A68_JOB)) {
-      rewrite_script_source ();
-    }
-#endif
-    if (OPTION_DEBUG (&A68_JOB)) {
-      state_license (STDOUT_FILENO);
-    }
-#if defined (BUILD_A68_COMPILER)
-    if (OPTION_OPT_LEVEL (&A68_JOB) > 0) {
-      char libname[BUFFER_SIZE];
-      void *a68_lib;
-      struct stat srcstat, objstat;
-      int ret;
-      announce_phase ("dynamic linker");
-      ASSERT (snprintf (libname, SNPRINTF_SIZE, "%s", FILE_LIBRARY_NAME (&A68_JOB)) >= 0);
-// Check whether we are doing something rash.
-      ret = stat (FILE_SOURCE_NAME (&A68_JOB), &srcstat);
-      ABEND (ret != 0, ERROR_ACTION, FILE_SOURCE_NAME (&A68_JOB));
-      ret = stat (libname, &objstat);
-      ABEND (ret != 0, ERROR_ACTION, libname);
-      if (OPTION_RERUN (&A68_JOB)) {
-        ABEND (ST_MTIME (&srcstat) > ST_MTIME (&objstat), "library outdates source", "cannot RERUN");
-      }
-// First load a68g itself so compiler code can resolve a68g symbols.
-      a68_lib = dlopen (NULL, RTLD_NOW | RTLD_GLOBAL);
-      ABEND (a68_lib == NULL, ERROR_RESOLVE, dlerror ());
-// Then load compiler code.
-      compile_lib = dlopen (libname, RTLD_NOW | RTLD_GLOBAL);
-      ABEND (compile_lib == NULL, ERROR_RESOLVE, dlerror ());
-    } else {
-      compile_lib = NULL;
-    }
+  if (OPTION_DEBUG (&A68_JOB)) {
+    state_license (A68_STDOUT);
+  }
+  if (ERROR_COUNT (&A68_JOB) == 0 && OPTION_COMPILE (&A68_JOB) == A68_FALSE && 
+     (OPTION_CHECK_ONLY (&A68_JOB) ? OPTION_RUN (&A68_JOB) : A68_TRUE)) {
     announce_phase ("genie");
-    genie (compile_lib);
-// Unload compiler library.
-    if (OPTION_OPT_LEVEL (&A68_JOB) > 0) {
-      int ret = dlclose (compile_lib);
-      ABEND (ret != 0, ERROR_ACTION, dlerror ());
-    }
+    GetRNGstate ();
+    A68 (f_entry) = TOP_NODE (&A68_JOB);
+    A68 (close_tty_on_exit) = A68_FALSE;
+#if defined (BUILD_A68_COMPILER)
+    plugin_driver_genie ();
 #else
-    announce_phase ("genie");
-    genie (NO_NODE);
+    genie ((void *) NULL);
 #endif
 // Free heap allocated by genie.
     genie_free (TOP_NODE (&A68_JOB));
@@ -517,8 +495,8 @@ void compiler_interpreter (void)
 // Normal end of program.
     diagnostics_to_terminal (TOP_LINE (&A68_JOB), A68_RUNTIME_ERROR);
     if (OPTION_DEBUG (&A68_JOB) || OPTION_TRACE (&A68_JOB) || OPTION_CLOCK (&A68_JOB)) {
-      ASSERT (snprintf (A68 (output_line), SNPRINTF_SIZE, "\nGenie finished in %.2f seconds\n", seconds () - A68 (cputime_0)) >= 0);
-      WRITE (STDOUT_FILENO, A68 (output_line));
+      ASSERT (a68_bufprt (A68 (output_line), SNPRINTF_SIZE, "\nGenie finished in %.2f seconds\n", seconds () - A68 (cputime_0)) >= 0);
+      WRITE (A68_STDOUT, A68 (output_line));
     }
     verbosity ();
   }
@@ -548,40 +526,9 @@ void compiler_interpreter (void)
 // Cleaning up the intermediate files.
 #if defined (BUILD_A68_COMPILER)
   announce_phase ("clean up intermediate files");
-  if (OPTION_OPT_LEVEL (&A68_JOB) >= OPTIMISE_0 && OPTION_REGRESSION_TEST (&A68_JOB) && !OPTION_KEEP (&A68_JOB)) {
-    if (emitted) {
-      a68_rm (FILE_OBJECT_NAME (&A68_JOB));
-    }
-    a68_rm (FILE_LIBRARY_NAME (&A68_JOB));
-  }
-  if (OPTION_RUN_SCRIPT (&A68_JOB) && !OPTION_KEEP (&A68_JOB)) {
-    if (emitted) {
-      a68_rm (FILE_OBJECT_NAME (&A68_JOB));
-    }
-    a68_rm (FILE_SOURCE_NAME (&A68_JOB));
-    a68_rm (FILE_LIBRARY_NAME (&A68_JOB));
-  } else if (OPTION_COMPILE (&A68_JOB)) {
-    build_script ();
-    if (!OPTION_KEEP (&A68_JOB)) {
-      if (emitted) {
-        a68_rm (FILE_OBJECT_NAME (&A68_JOB));
-      }
-      a68_rm (FILE_LIBRARY_NAME (&A68_JOB));
-    }
-  } else if (OPTION_OPT_LEVEL (&A68_JOB) == OPTIMISE_0 && !OPTION_KEEP (&A68_JOB)) {
-    if (emitted) {
-      a68_rm (FILE_OBJECT_NAME (&A68_JOB));
-    }
-    a68_rm (FILE_LIBRARY_NAME (&A68_JOB));
-  } else if (OPTION_OPT_LEVEL (&A68_JOB) > OPTIMISE_0 && !OPTION_KEEP (&A68_JOB)) {
-    if (emitted) {
-      a68_rm (FILE_OBJECT_NAME (&A68_JOB));
-    }
-  } else if (OPTION_RERUN (&A68_JOB) && !OPTION_KEEP (&A68_JOB)) {
-    if (emitted) {
-      a68_rm (FILE_OBJECT_NAME (&A68_JOB));
-    }
-  }
+  plugin_driver_clean (emitted);
+#else
+  (void) emitted;
 #endif
 }
 
@@ -612,21 +559,18 @@ void a68_exit (int code)
   free_option_list (OPTION_LIST (&A68_JOB));
   a68_free (A68 (node_register));
   a68_free (A68 (options));
-//
   discard_heap ();
-//
   a68_free (FILE_PATH (&A68_JOB));
   a68_free (FILE_INITIAL_NAME (&A68_JOB));
   a68_free (FILE_GENERIC_NAME (&A68_JOB));
   a68_free (FILE_SOURCE_NAME (&A68_JOB));
   a68_free (FILE_LISTING_NAME (&A68_JOB));
   a68_free (FILE_OBJECT_NAME (&A68_JOB));
-  a68_free (FILE_LIBRARY_NAME (&A68_JOB));
+  a68_free (FILE_PLUGIN_NAME (&A68_JOB));
   a68_free (FILE_BINARY_NAME (&A68_JOB));
   a68_free (FILE_PRETTY_NAME (&A68_JOB));
   a68_free (FILE_SCRIPT_NAME (&A68_JOB));
   a68_free (FILE_DIAGS_NAME (&A68_JOB));
-//
   a68_free (A68_MP (mp_one));
   a68_free (A68_MP (mp_pi));
   a68_free (A68_MP (mp_half_pi));
@@ -636,7 +580,6 @@ void a68_exit (int code)
   a68_free (A68_MP (mp_ln_pi));
   a68_free (A68_MP (mp_180_over_pi));
   a68_free (A68_MP (mp_pi_over_180));
-//
   exit (code);
 }
 
@@ -650,9 +593,8 @@ int main (int argc, char *argv[])
   A68 (close_tty_on_exit) = A68_TRUE;
   FILE_DIAGS_FD (&A68_JOB) = -1;
 // Get command name and discard path.
-  bufcpy (A68 (a68_cmd_name), argv[0], BUFFER_SIZE);
-  int k;
-  for (k = (int) strlen (A68 (a68_cmd_name)) - 1; k >= 0; k--) {
+  a68_bufcpy (A68 (a68_cmd_name), argv[0], BUFFER_SIZE);
+  for (int k = (int) strlen (A68 (a68_cmd_name)) - 1; k >= 0; k--) {
 #if defined (BUILD_WIN32)
     char delim = '\\';
 #else
@@ -700,7 +642,7 @@ int main (int argc, char *argv[])
     FILE_SOURCE_NAME (&A68_JOB) = NO_TEXT;
     FILE_LISTING_NAME (&A68_JOB) = NO_TEXT;
     FILE_OBJECT_NAME (&A68_JOB) = NO_TEXT;
-    FILE_LIBRARY_NAME (&A68_JOB) = NO_TEXT;
+    FILE_PLUGIN_NAME (&A68_JOB) = NO_TEXT;
     FILE_BINARY_NAME (&A68_JOB) = NO_TEXT;
     FILE_PRETTY_NAME (&A68_JOB) = NO_TEXT;
     FILE_SCRIPT_NAME (&A68_JOB) = NO_TEXT;
@@ -710,23 +652,22 @@ int main (int argc, char *argv[])
     read_env_options ();
 // Posix copies arguments from the command line.
     if (argc <= 1) {
-      online_help (STDOUT_FILENO);
+      online_help (A68_STDOUT);
       a68_exit (EXIT_FAILURE);
     }
-    int argcc;
-    for (argcc = 1; argcc < argc; argcc++) {
-      add_option_list (&(OPTION_LIST (&A68_JOB)), argv[argcc], NO_LINE);
+    for (int k = 1; k < argc; k++) {
+      add_option_list (&(OPTION_LIST (&A68_JOB)), argv[k], NO_LINE);
     }
     if (!set_options (OPTION_LIST (&A68_JOB), A68_TRUE)) {
       a68_exit (EXIT_FAILURE);
     }
 // State license.
     if (OPTION_LICENSE (&A68_JOB)) {
-      state_license (STDOUT_FILENO);
+      state_license (A68_STDOUT);
     }
 // State version.
     if (OPTION_VERSION (&A68_JOB)) {
-      state_version (STDOUT_FILENO);
+      state_version (A68_STDOUT);
     }
 // Start the UI.
     init_before_tokeniser ();
